#include "StdAfx.h"
#include "TSPalgorithm.h"
#include <algorithm>

const int rand_value = 100;

TSPalgorithm::TSPalgorithm(void)
{
}


TSPalgorithm::~TSPalgorithm(void)
{
}


// Set up TSP problem to run
void TSPalgorithm::Initialize( const int& iter,
							   CoordMatrix* mat,
							   const float& temp,
							   const float& cool )
{
	// Set the matrix pointer for Tour object
	tour.SetMatrix( mat );
	iterations = iter;
	temperature = temp;
	cooling_rate = cool;
}


// Run the optimization algorithm
void TSPalgorithm::Run()
{	
	tour.Reset();	
	tour.CreateRandomTour();
	//tour.CreateNearestNeighbourTour();
	size = tour.TourSize();
	
	// Create the optimal tour
	double best = 99999999;

	for ( int i = 0; i < iterations; i++ )
	{
		Rotate( i );
		Reverse( i );		
		SwapRandomPair( i );	

		double dist = tour.TourDistance();

		if ( dist < best )
		{
			best = dist;
			Notify( (const int&) tour.TourDistance(), (const int&) i );
		}
	}

	Notify( (const int&) best, iterations - 1 );
}


// Get the tour node, given the index
int TSPalgorithm::GetTourCity( const int& index )
{
	int node = tour.GetCity( index );
	return node;
}


// Notify observers of changes
void TSPalgorithm::Notify( const int& dist, const int& iter )  
{  
    for ( std::vector<Observer*>::iterator it = ob_set.begin();  
          it != ob_set.end();  
          it++ )  
    {  
		Observer* ob = *it;  
		ob->UpdateDisplay( dist, iter );  
    }  
}  


// Add observer
void TSPalgorithm::AddObserver( Observer* ob )
{
	ob_set.push_back( ob );
}

// Do 2-opt combinations
void TSPalgorithm::TwoOpt( const int& iter )
{
	double dist = tour.TourDistance();	
	
	// Get feasible end points
	int c1_min = 1;
	int c1_max = size - 5;
	int c1;
	
	do
	{
		c1 = rand() % ( size );
	}
	while ( c1 < c1_min || c1 > c1_max );
		
	int c2 = c1 + 1;
	int c3_min = c1 + 3;
	int c3 = c3_min + rand() % ( size - 1 - c3_min );
	int c4 = c3 + 1;

	// Store the city values
	int c1s = tour.GetCity( c1 );
	int c2s = tour.GetCity( c2 );
	int c3s = tour.GetCity( c3 );
	int c4s = tour.GetCity( c4 );
	
	tour.DoTwoOpt( c1, c2, c3, c4 );						
	
	double new_dist = tour.TourDistance();

	bool accept = Accept( new_dist, dist );

	if ( !accept )
	{		
		//tour.SetCities( store );	
		tour.SetCity( c1, c1s );
		tour.SetCity( c2, c2s );
		tour.SetCity( c3, c3s );
		tour.SetCity( c4, c4s );

		dist = tour.TourDistance();	
	}		
}

// Do 3-opt combinations
void TSPalgorithm::ThreeOpt( const int& iter )
{
	// Get tour size
	int size = tour.TourSize();

	for ( int i = 0; i < size - 5; i++ )
	{
		for ( int j = i + 3; j < size - 3; j++ )
		{
			for ( int k = j + 3; k < size - 1; k++ )
			{
				int c1 = i;
				int c2 = i + 1;
				int c3 = j;
				int c4 = j + 1;
				int c5 = k;
				int c6 = k + 1;

				tour.DoThreeOpt( c1, c2, c3, c4, c5, c6 );	
			}
		}	

		Notify( (const int&) tour.TourDistance(), (const int&) iter );
	}
}


// Perform rearrangement of chosen cities:
// choose section of path and reverse the order
void TSPalgorithm::Reverse( const int& iter )
{
	// Store the tour	
	std::vector<int> store;
	tour.Copy( store );

	double dist = tour.TourDistance();
	
	int c1 = rand() % ( size );

	int c2;

	do
	{
		c2 = rand() % ( size );
	}
	while ( c1 == c2  );

	tour.ReverseCities( c1, c2 );

	double new_dist = tour.TourDistance();

	bool accept = Accept( new_dist, dist );

	if ( !accept )
	{		
		tour.SetCities( store );	
	}		
}


// Perform rearrangement of chosen cities:
// choose sections of path and rotate the order
void TSPalgorithm::Rotate( const int& iter )
{
	// Store the tour	
	std::vector<int> store;
	tour.Copy( store );

	double dist = tour.TourDistance();	

	int c[ 3 ];

	c[ 0 ] = rand() % ( size ); 
	c[ 1 ] = rand() % ( size ); 
	c[ 2 ] = rand() % ( size ); 

	while ( c[ 0 ] == c[ 1 ] || 
		    c[ 0 ] == c[ 2 ] || 
			c[ 1 ] == c[ 2 ] )
	{
		c[ 0 ] = rand() % ( size ); 
		c[ 1 ] = rand() % ( size ); 
		c[ 2 ] = rand() % ( size ); 
	}
	
	std::sort( c, c + 3 );
	tour.RotateCities( c[ 0 ], c[ 1 ], c[ 2 ] );

	double new_dist = tour.TourDistance();

	bool accept = Accept( new_dist, dist );

	if ( !accept )
	{		
		tour.SetCities( store );	
	}	
	else
	{
		if ( new_dist < dist ) 
		{
			Notify( (const int&) new_dist, (const int&) iter );
		}
	}
}


// Do pairwise swap operator on random selected city pairs
void TSPalgorithm::SwapRandomPair( const int& iter )
{
	int c1 = 1 + rand() % ( size - 1 );		
	int c2;
		
	do 
	{
		c2 = 1 + rand() % ( size - 2 );		
	}
	while ( c1 == c2 );

	double dist = tour.TourDistance();	
	
	int tmp1 = tour.GetCity( c1 );
	int tmp2 = tour.GetCity( c2);

	tour.SetCity( c1, tmp2 );
	tour.SetCity( c2, tmp1 );

	double new_dist = tour.TourDistance();

	bool accept = Accept( new_dist, dist );

	if ( !accept )
	{		
		tour.SetCity( c1, tmp1 );	
		tour.SetCity( c2, tmp2 );	
	}	
	else
	{
		if ( new_dist < dist ) 
		{
			Notify( (const int&) new_dist, (const int&) iter );
		}
	}
}


// Do pairwise swap operator on all selected city pairs
void TSPalgorithm::RepairAllPairs()
{
	for ( int i = 1; i < size - 4; i++ )
	{
		for ( int j = i + 3; j < size - 1; j++ )
		{
			int c1 = i;		
			int c2 = i + 1;
			int c3 = j;
			int c4 = j + 1;
		
			double dist = tour.TourDistance();	
	
			int tmp1 = tour.GetCity( c1 );
			int tmp2 = tour.GetCity( c2);
			int tmp3 = tour.GetCity( c3 );
			int tmp4 = tour.GetCity( c4);

			tour.DoTwoOpt( c1, c2, c3, c4 );		

			double new_dist = tour.TourDistance();

			bool accept = Accept( new_dist, dist );

			if ( !accept )
			{		
				tour.SetCity( c1, tmp1 );	
				tour.SetCity( c2, tmp2 );	
				tour.SetCity( c3, tmp3 );	
				tour.SetCity( c4, tmp4 );	
			}
			else
			{
				int p = 0;
			}
		}
	}

}



// Returns true if we should accept the new solution
// false otherwise.  Use standard thermodynamics equation
// P = exp( -(old-new) / kt ) where k = Boltzmann's constant
bool TSPalgorithm::Accept( const double& new_dist,
						   const double& old_dist )
{
	// Calculate acceptance probability: declines with time
	double diff = (double) -1.0 * (new_dist - old_dist);
	temperature = temperature * cooling_rate;
	
	// Calculate probability of accepting worse solutions
	// This probability declines with time
	double P = exp( diff / temperature );

	if ( P > 1.0 ) return true;
	else
	{
		// Generate random number between 0 and 1
		double p = rand() / double(RAND_MAX);

		return p < P;
	}
}


